/**
 * moiforth - moi's forth interpreter.
 *
 * @file main.c
 */



#include "token.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>



/**
 * O-̔Ԓn̑g
 */
typedef struct name_instance_pair {
    char name_pointer;              /**< Oo^Ăʒu */
    char instance_pointer;          /**< ̂o^Ăʒu */
} NAME_INSTANCE_PAIR;



void push(long val);
long pop(void);
void push_branch_stack(char adrs);
unsigned char pop_branch_stack(void);
void proc_prog(void);
void init(void);
char reg_word_name(char *str);
int isnumber(char *str);
void assoc_word_name(char *name);
char get_word_address(char *name);
char reg_var_name(char *str);
void assoc_var_name(char *name);
char get_var_address(char *name);
void parse(SRC_BUFFER *cmd);
void load_prog(char *fname, SRC_BUFFER *cmd);
int main(int argc, char *argv[]);
char *get_token(SRC_BUFFER *srcbuf);
char *get_toquote(SRC_BUFFER *srcbuf);

void call_push(void);
void plus(void);
void print(void);
void word_if(void);
void word_else(void);
void word_then(void);
void call_do(void);
void call_loop(void);
void call_i(void);
void call_cr(void);
void call_word(void);
void proc_ret(void);
void call_dotquote(void);
void call_varadr(void);



/**
 * z}Ṽ}VR[h
 */
enum CODE_ENUM {
    CODE_END,
    CODE_PUSH,
    CODE_PLUS,
    CODE_PRINT,
    CODE_IF,

    CODE_ELSE,
    CODE_THEN,
    CODE_DO,
    CODE_LOOP,
    CODE_I,

    CODE_CR,
    CODE_CALL,
    CODE_RET,
    CODE_DOTQUOTE,
    CODE_VARADR
};



#define WORD_HEAP_SIZE (sizeof(word_heap) / sizeof(char))       /**< [ho^̈̃TCY */



#define PS(s) printf((s));      /**< fobOp */
#define PN(s,n) printf((s),(n));        /**< fobOp */



const char *sep = " \t\n\xd\xa";    /**< \͂̋؂q */



char stack[100];                /**< X^bN */
int pos;                        /**< X^bN̈ʒu|C^ */

unsigned char prog[100];        /**< vOi[̈ */
int prog_cnt;                   /**< vOJE^ */
int base_prog_cnt;              /**< vOJE^̏ʒu */

char branch_stack[10];          /**< X^bN */
int branch_pos;                 /**< X^bN̈ʒu|C^ */

long loop_cnt_stack[10];        /**< [vJE^pX^bN */
long loop_end_stack[10];        /**< [vIlpX^bN */
int loop_pos;                   /**< [vX^bN̈ʒu|C^ */

char word_heap[100];            /**< [ho^̈ */
unsigned char word_heap_index;  /**< [ho^̖̈gp̈̐擪w */

NAME_INSTANCE_PAIR word_proc[100];  /**< [hǗe[u */
unsigned char word_proc_head;       /**< [hǗe[u̖gpʒu̐擪 */

unsigned char var[100];             /**< ϐ̂̊i[̈ */
int var_pos;                        /**< ϐ̗̖̈gp̈̐擪w */

char varname_heap[100];             /**< ϐo^̈ */
unsigned char varname_heap_index;   /**< ϐo^̖̈gp̈̐擪w */

NAME_INSTANCE_PAIR var_table[100];  /**< ϐǗe[u */
unsigned char var_table_head;       /**< ϐǗe[u̖gpʒu̐擪 */

char word_first_pos;            /**< [ho^̐擪Ԓn */

char quit_enabled;              /**< Θb甲oƂ1 */
char proc_disable_cnt;          /**< 1ȏ̂ƂAvvg̎s֎~ */
int parse_idx;                  /**< \͂ł̃vOʒu */

void (*jump_table[])(void) = {  /**< [h̃Wve[u */
    NULL, call_push, plus, print, word_if, word_else, word_then, call_do, call_loop, call_i,
    call_cr, call_word, proc_ret, call_dotquote, call_varadr
};



/** 
 * lX^bNɐς
 */
void push(long val)
{
    pos = pos - 4;
    *(long *)(stack + pos) = val;
}



/**
 * X^bNlo
 */
long pop(void)
{
    long val = *(long *)(stack + pos);

    pos += 4;

    return val;
}



/** 
 * X^bNɔԒnς
 */
void push_branch_stack(char adrs)
{
    branch_pos--;
    branch_stack[branch_pos] = adrs;
}



/** 
 * X^bNԒno
 */
unsigned char pop_branch_stack(void)
{
    return branch_stack[branch_pos++];
}



/** 
 * Z
 */
void plus(void)
{
    long a, b, c;

    a = pop();
    b = pop();

    c = a + b;

    push(c);
}



/** 
 * \
 */
void print(void)
{
    long a;

    a = pop();
    printf("%ld", a);
}



/** 
 * [h'do's
 */
void call_do(void)
{
    loop_pos--;

    /* X^bN̓e[vX^bNɓ */
    loop_cnt_stack[loop_pos] = pop();
    loop_end_stack[loop_pos] = pop();
}



/** 
 * [h'loop's
 */
void call_loop(void)
{
    loop_cnt_stack[loop_pos]++;
    if (loop_cnt_stack[loop_pos] >= loop_end_stack[loop_pos]) {
        /* [v𔲂 */
        loop_pos++;
        prog_cnt++;

    } else {
        /* [v擪ɖ߂ */
        prog_cnt = prog[prog_cnt];

    }
}



/** 
 * [h'i's
 */
void call_i(void)
{
    /* X^bNɈԓ̃[vJE^push */
    push(loop_cnt_stack[loop_pos]);
}



/** 
 * [h'cr's
 */
void call_cr(void)
{
    printf("\n");
}



/** 
 * vOs
 */
void proc_prog(void)
{
    prog_cnt = base_prog_cnt;

    while (prog[prog_cnt] != CODE_END) {
        (*jump_table[prog[prog_cnt++]])();
    }
}



/**
 * X^bNɃf[^PUSH
 */
void call_push(void)
{
    push(*(long *)(prog + prog_cnt));
    prog_cnt += 4;
}



/**
 * X^bÑf[^ɉď𕪊򂷂
 */
void word_if(void)
{
    if (pop()) {
        prog_cnt++;

    } else {
        prog_cnt = prog[prog_cnt];

    }
}



/**
 * elses
 */
void word_else(void)
{
    prog_cnt = prog[prog_cnt];
}



/**
 * thens
 */
void word_then(void)
{
}



/**
 * [U`[hĂяo
 */
void call_word(void)
{
    push_branch_stack(prog_cnt);
    prog_cnt = prog[prog_cnt];
}



/**
 * [U`[h̃^[s
 */
void proc_ret(void)
{
    prog_cnt = pop_branch_stack();
}



/**
 * ."s
 */
void call_dotquote(void)
{
    unsigned short j;
    unsigned short len = *(unsigned short *)(prog + prog_cnt);

    prog_cnt += 2;
    for (j = 0; j <len ; j++) {
        printf("%c", prog[prog_cnt++]);
    }
}



/**
 * ϐ̔ԒnX^bNɐς
 */
void call_varadr(void)
{
    push(prog[prog_cnt]);
    prog_cnt++;
}



/** 
 * 
 */
void init(void)
{
    int j;

    for (j = 0; j < sizeof(prog) / sizeof(char); j++) {
        prog[j] = CODE_END;
    }

    prog_cnt = 0;
    base_prog_cnt = 0;
    word_heap_index = 0;
    word_proc_head = 0;
    
    var_pos = 0;
    varname_heap_index = 0;
    var_table_head = 0;

    pos = sizeof(stack) / sizeof(char);
    branch_pos = sizeof(branch_stack) / sizeof(char);
    loop_pos = sizeof(loop_cnt_stack) / sizeof(char);

    quit_enabled = 0;
    proc_disable_cnt = 0;
    parse_idx = 0;
}



/** 
 * [ho^.
 *
 * @param str
 * @return o^̃q[v̐擪ʒu
 */
char reg_word_name(char *str)
{
    char reg_pos = word_heap_index;

    /* Rs[ */
    while (*str != '\x0') {
        word_heap[word_heap_index++] = *str++;
    }

    /* I[ */
    word_heap[word_heap_index++] = '\x0';

    /* o^̐擪ʒuԂ */
    return reg_pos;
}



/** 
 * 񂪐̂Ƃ1Ԃ
 */
int isnumber(char *str)
{
    while (*str != '\x0') {
        if (!isdigit(*str)) {
            return 0;
        }
        str++;
    }

    return 1;
}



/**
 * [hƏԒnƂ֘At
 * @param[in] name o^[h
 */
void assoc_word_name(char *name)
{
    /* [hq[v̈ɓo^ */
    word_proc[word_proc_head].name_pointer = reg_word_name(name);

    /* ̎_̃vOʒuo^([h̊JnʒuɂȂ) */
    word_proc[word_proc_head].instance_pointer = prog_cnt;

    /* gĂȂo^e[u擪ɂ */
    word_proc_head++;
}



/**
 * Ŏw肵[h̃vOʒuԂ
 */
char get_word_address(char *name)
{
    unsigned char j;

    for (j = 0; j < 100; j++) {
        unsigned char heap_index = word_proc[j].name_pointer;

        if (strcmp(&word_heap[heap_index], name) == 0) {
            return j;
        }
    }

    return -1;
}



/** 
 * ϐo^.
 *
 * @param str
 * @return o^̃q[v̐擪ʒu
 */
char reg_var_name(char *str)
{
    char reg_pos = varname_heap_index;

    /* Rs[ */
    while (*str != '\x0') {
        varname_heap[varname_heap_index++] = *str++;
    }

    /* I[ */
    varname_heap[varname_heap_index++] = '\x0';

    /* o^̐擪ʒuԂ */
    return reg_pos;
}



/**
 * ϐƎ̔ԒnƂ֘At
 */
void assoc_var_name(char *name)
{
    /* ϐq[v̈ɓo^ */
    var_table[var_table_head].name_pointer = reg_var_name(name);

    /* ϐ̊蓖ĔԒno^ */
    var_table[var_table_head].instance_pointer = var_pos;
    var_pos += 2;

    /* gĂȂo^e[u擪ɂ */
    var_table_head++;
}



/**
 * Ŏw肵ϐ̎̔ԒnԂ
 */
char get_var_address(char *name)
{
    unsigned char j;

    for (j = 0; j < 100; j++) {
        unsigned char heap_index = var_table[j].name_pointer;

        if (strcmp(&varname_heap[heap_index], name) == 0) {
            return var_table[j].instance_pointer;
        }
    }

    return -1;
}



/** 
 * \͂
 */
void parse(SRC_BUFFER *cmd)
{
    char *token;
    char adrs;

    if (!proc_disable_cnt) {
        parse_idx = 0;
    }

    token = get_token(cmd);
    while (*token != '\x0') {
        if (strcmp(token, "+") == 0) {
            prog[parse_idx] = CODE_PLUS;
            parse_idx++;

        } else if (strcmp(token, ".") == 0) {
            prog[parse_idx] = CODE_PRINT;
            parse_idx++;

        } else if (isnumber(token)) {
            prog[parse_idx] = CODE_PUSH;
            *(long *)(prog + parse_idx + 1) = atol(token);
            parse_idx += 5;

        } else if (strcmp(token, "if") == 0) {
            prog[parse_idx] = CODE_IF;
            push_branch_stack(parse_idx + 1);
            parse_idx += 2;
            proc_disable_cnt++;

        } else if (strcmp(token, "else") == 0) {
            prog[parse_idx] = CODE_ELSE;
            prog[pop_branch_stack()] = parse_idx + 2;
            push_branch_stack(parse_idx + 1);
            parse_idx += 2;

        } else if (strcmp(token, "then") == 0) {
            prog[parse_idx] = CODE_THEN;
            prog[pop_branch_stack()] = parse_idx + 1;
            parse_idx++;
            proc_disable_cnt--;

        } else if (strcmp(token, "do") == 0) {
            prog[parse_idx] = CODE_DO;
            push_branch_stack(parse_idx + 1);
            parse_idx++;
            proc_disable_cnt++;

        } else if (strcmp(token, "loop") == 0) {
            prog[parse_idx] = CODE_LOOP;
            prog[parse_idx + 1] = pop_branch_stack();
            parse_idx += 2;
            proc_disable_cnt--;

        } else if (strcmp(token, "i") == 0) {
            prog[parse_idx] = CODE_I;
            parse_idx++;

        } else if (strcmp(token, "cr") == 0) {
            prog[parse_idx] = CODE_CR;
            parse_idx++;

        } else if (strcmp(token, ":") == 0) {
            /* :̎̒P[hƂ */
            token = get_token(NULL);
            assoc_word_name(token);

        } else if (strcmp(token, ";") == 0) {
            /* [hI[R[hǉ */
            prog[parse_idx] = CODE_RET;
            parse_idx++;
            base_prog_cnt = parse_idx;

        } else if (strcmp(token, ".\"") == 0) {
            int j;

            prog[parse_idx++] = CODE_DOTQUOTE;
            token = get_toquote(NULL);
            *(unsigned short *)(prog + parse_idx) = strlen(token);
            parse_idx += 2;
            for (j = 0; token[j] != '\x0'; j++) {
                prog[parse_idx++] = token[j];
            }

        } else if (strcmp(token, "variable") == 0) {
            /* g[NϐƂ */
            token = get_token(NULL);
            assoc_var_name(token);

        } else if ((adrs = get_var_address(token)) >= 0) {
            /* ϐ̔ԒnX^bNPUSH */
            prog[parse_idx] = CODE_VARADR;
            prog[parse_idx + 1] = adrs;
            parse_idx += 2;

        } else if ((adrs = get_word_address(token)) >= 0) {
            /* [U`[hĂяo𖄂ߍ */
            prog[parse_idx] = CODE_CALL;
            prog[parse_idx + 1] = adrs;
            parse_idx += 2;

        } else {
            printf("undefined word: %s\n", token);
            exit(1);

        }

        token = get_token(NULL);
    }

    prog[parse_idx] = CODE_END;
}



/**
 * ForthŏꂽvO[h
 */
void load_prog(char *fname, SRC_BUFFER *cmd)
{
    FILE *fp;
    unsigned int idx = 0;
    char c;

    fp = fopen(fname, "r");

    while ((c = fgetc(fp)) != EOF) {
        cmd->buf[idx++] = c;
    }

    cmd->buf[idx] = '\x0';
    cmd->pos = 0;

    fclose(fp);
}



int main(int argc, char *argv[])
{
    char buf[100];
    SRC_BUFFER cmd_buf = {buf, 0};

    init();

    if (argc >= 2) {
        load_prog(argv[1], &cmd_buf);
        parse(&cmd_buf);

        proc_prog();

    } else {
        while (!quit_enabled) {
            printf(">");
            scanf("%100[^\n]%*[^\n]", cmd_buf.buf);
            getchar();

            parse(&cmd_buf);
            if (!proc_disable_cnt) {
                proc_prog();
            }
        }

    }

    return 0;
}




